////////////////////////////////////////////////////////////////////////////////////////////////
//
// Kalendar a Javascript calendar widget.
//
// Written by: Don Dugger
//       Date: 2024-04-29
//
// Copyright (C) 2024 Acme Software Works, Inc.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
// GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
// OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
////////////////////////////////////////////////////////////////////////////////////////////////
/** 
 * This JavaScript file contains the Kalendar class
 * 
 * @projectname Kalendar
 * @version 1.o
 * @author Don Dugger
 * @copyright 2024
 * 
 */
/** @class */
class Kalendar {

    /** Holidays and Events
     *```none
     *Event structure "mm/dd/T:text"
     *    mm              month 1-12
     *    dd or position  day of month ( 1-31 or day position )
     *    T               Event type "H" = Holiday, "I" = Information, "P" = Personal
     *    text            Displayed texmt
     *```
     *<p>  
     *   The day position:<br />
     *    First 3 characters are the day of the week.
     *    Followed by the position first, second, third, fourth or last.
     *    The day position ignores case.
     *</p>
     */
    static events = ["01/1/H:New Year's Day",
                     "02/02/H:Groundhog Day",
                     "02/14/H:Valentine's Day",
                     "02/MonThird/H:President's Day",
                     "03/05/H:Mother-in-Law Day",
                     "03/SunSecond/I:Daylight Savings Starts",
                     "03/17/H:St. Patrick's Day",
                     "04/01/H:April Fool's Day",
                     "04/15/H:Income Tax Day",
                     "04/28/H:Arbor Day",
                     "05/1/H:International Workers’ Day",
                     "05/SunSecond/H:Mother's Day",
                     "05/SatThird/H:Armed Forces Day",
                     "05/MonLast/H:Memorial Day",
                     "06/SunThird/H:Father's Day",
                     "07/04/H:Independence Day",
                     "09/MonFirst/H:Labor Day",
                     "10/MonSecond/H:Indigenous People's Day",
                     "10/31/H:Halloween",
                     "11/SunFirst/I:Daylight Savings Ends",
                     "11/11/H:Veterans' Day",
                     "11/ThuFourth/H:Thanksgiving Day",
                     "12/24/H:Christmas Eve",
                     "12/25/H:Christmas",
                     "12/31/H:New Year's Eve"];
    //
    //
    static notesList = [];
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     * Initialize
     * @param {string} div - This is the <div> where the calendar will be displayed. 
     */
    static init(div) {
        this.date = new Date(); // Temporary for testing
        this.season = Date.getSeasons();
        this.date.setHours( 0, 0, 0, 0 );
        this.year = Kalendar.date.getFullYear();
        this.month = this.date.getMonth();
        Kalendar.getOffset();
        this.days_in_month = Kalendar.getDaysInMonth();
        Kalendar.getPrevNextMonth();
        this.div = div;
        this.tab = "kal-table";
        this.head = "kal-header";
        this.displayed = false;
        /** If the browser is given permission to obtain the location it will be set automatically.
         Other wise it can be set with the setLocation() function.
         Note the call to display() is in navLocation()
         */
        if (navigator.geolocation) {
            navigator.geolocation.getCurrentPosition(Kalendar.navLocation,null,{enableHighAccuracy: true});
        } else {
            this.latitude = NaN;
            this.longitude = NaN;
        }
        Kalendar.display();
        this.notesAllowed = false;
    } // End of init()
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     * The constructor
     * @constructor
     * @param {string} div -This is the <div> where the calendar will be displayed. 
     */
    constructor(div) {
        Kalendar.init(div);
    } // End of constructor()
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     *  set the location's latitude and longitude for sunrise and sunset.
     *       (If not set the sunrise & sunset will not be in the tool tip.)
     *  @param {number} latitude     The latitude for the sun rise/set times     
     *  @param {number} longitude    The longitude for the sun rise/set times     
     */
    static setLocation(latitude, longitude) {
        this.latitude = latitude;
        this.longitude = longitude;
        Kalendar.display();
    } // End of setLocation()
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     *  The navigator.geolocation read-only property returns a Geolocation object that gives
     *  Web content access to the location of the device. This allows a website or app to offer
     *  customized results based on the user's location. 
     *
     *  @param {position} pos - The location returned by the browser
     */
    static navLocation(pos) {
        Kalendar.latitude = pos.coords.latitude;
        Kalendar.longitude = pos.coords.longitude;
        Kalendar.display();
    } // End of navLocation()
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     * Set the date, year and mouth to be displayed.
     *
     * @param {number} year  - The year to be set
     * @param {number} month - The month to be set
     */
    static setDate(year,month) {
        this.date = new Date(year,month-1,1);
        this.season = Date.getSeasons();
        this.date.setHours( 0, 0, 0, 0 );
        this.year = this.date.getFullYear();
        this.month = this.date.getMonth();
        this.getOffset();
        this.days_in_month = this.getDaysInMonth();
        this.getPrevNextMonth();
        Kalendar.display();
    } // End of setDate()
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     *  Get the offset of the first row to the 1st date. In order to determine the first cell
     *  to be populated.
     *
     *  @return {number} - The offset
     */
    static getOffset() {
        const tmp_date = new Date(this.year,this.month,1);
        this.offset = tmp_date.getDay();
        return this.offset;
    } // End of getOffset()
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /** 
     *  Get total number of days in the current month.
     *
     * @param {number} [year=this.year]   - The year to be used
     * @param {number} [month=this.month] - The month to get the number
     */
    static getDaysInMonth(year = this.year, month = this.month) {
        return new Date(year, month+1, 0).getDate();
    } // End of getDaysInMonth()
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     *  Get the month before and after the current month and the number of days in the previous
     *  month in oder to fill in the cells at the beginning of the calendar.
     */
    static getPrevNextMonth() {
        this.prev_date = new Date(this.year, this.month-1);
        this.next_date = new Date(this.year, this.month+1);
        this.prev_days_in_month = this.getDaysInMonth(this.prev_date.getFullYear(),this.prev_date.getMonth());

    } // End of getPrevNextMonth()
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     *  Go to the previous Year.
     */
    static prevYear() {
        this.setDate(this.year-1,this.month+1,1);
        this.display();
    } // End of prevYear()
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     *  Go to the next Year.
     */
    static nextYear() {
        this.setDate(this.year+1,this.month+1,1);
        this.display();
    } // End of nextYear()
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     *  Go to the previous month.
     */
    static prevMonth() {
        this.setDate(this.year,this.month,1);
        this.display();
    } // End of prevMonth()
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     *  Go to the next month.
     */
    static nextMonth() {
        this.setDate(this.year,this.month+2,1);
        this.display();
    } // End of nextMonth()
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     *  Add the the days of the week to heading of calendar table.
     */
    static dayNames = ['Sunday','Monday','Tuesday','Wednesday','Thursday','Friday','Saturday'];
    static addDayHeader() {
        var str = "<tr>\n\t";
        for (var day=0; day<7;++day) {
            str += "<td class=\"kalendar-head\" align='center'>" + this.dayNames[day] + "</td>";
        }
        str += "\n</tr>\n";
        $('#' + this.tab).append(str);
    } // End of addDayHeader()
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     *  Add the rows of calendar table.
     */
    /** @type {number} */
    static NEW_MOON      = 0;
    static FIRST_QUARTER = 1;
    static FULL_MOON     = 2;
    static LAST_QUARTER  = 3;
    static monthNames = ['January','February','March','April','May','June','July','August','September','October','November','December'];
    static addRows() {
        var ct = 0                  // Cell count
        var day_of_month = 0;       // Days in the current (will be set later)
        var day_of_next_month = 0;  // The count of days in the next month, this fills out the last cells.
        // Make sure there are enough rows to cover all the day in the current month.
        var num_rows = Math.ceil((this.offset + this.days_in_month)/7);
        // Init prev_day to the first cells day of the previous month.
        var prev_day = this.prev_days_in_month - this.offset + 1;
        /////////////////////////////////////////////////////////////////////////////////////////////
        // Get the phases of the Moon
        var phase_days = [];
        // Clear the moon phase array
        for ( var i=0; i<=this.days_in_month; ++i ) {
            phase_days[i] = -1;
        }
        /////////////////////////////////////////////////////////////////////////////////////////////
        // Initialize the moon phase object
        var moon = new MoonPhase()
        // Get the New Moon phases
        var phdate = moon.nextphase(this.year,this.month+1,1,0,0,this.NEW_MOON);
        phase_days[phdate.getDate()] = this.NEW_MOON;
        // Check for the second New Moon this month
        //      It should be noted that any given phase could appear twice in the same month.
        phdate = moon.nextphase(this.year,this.month+1,phdate.getDate()+2,0,0,this.NEW_MOON);
        if ( phdate.getMonth() == this.month ) {
            phase_days[phdate.getDate()] = this.NEW_MOON;
        }
        // Get the First Quarter phases
        phdate = moon.nextphase(this.year,this.month+1,1,0,0,this.FIRST_QUARTER);
        phase_days[phdate.getDate()] = this.FIRST_QUARTER;
        // Check for the second First Quarter this month
        phdate = moon.nextphase(this.year,this.month+1,phdate.getDate()+2,0,0,this.FIRST_QUARTER);
        if ( phdate.getMonth() == this.month ) {
            phase_days[phdate.getDate()] = this.FIRST_QUARTER;
        }
        // Get the Full Moon phases
        phdate = moon.nextphase(this.year,this.month+1,1,0,0,this.FULL_MOON);
        phase_days[phdate.getDate()] = this.FULL_MOON;
        // Check for the second Full Moon this month
        phdate = moon.nextphase(this.year,this.month+1,phdate.getDate()+2,0,0,this.FULL_MOON);
        if ( phdate.getMonth() == this.month ) {
            phase_days[phdate.getDate()] = this.FULL_MOON;
        }
        // Get the Last Quarter phases
        phdate = moon.nextphase(this.year,this.month+1,1,0,0,this.LAST_QUARTER);
        phase_days[phdate.getDate()] = this.LAST_QUARTER;
        // Check for the second Last Quarter this month
        phdate = moon.nextphase(this.year,this.month+1,phdate.getDate()+2,0,0,this.LAST_QUARTER);
        if ( phdate.getMonth() == this.month ) {
            phase_days[phdate.getDate()] = this.LAST_QUARTER;
        }
        ////////////////////////////////////////////////////////////////////////////////////////
        /**
         *  Put in the rows of the calendar.
         */
        for (var row=0; row<num_rows;++row) {
            // First the row tag
            var str = "<tr>\n\t";
            ////////////////////////////////////////////////////////////////////////////////////
            /**
             *  The Seven days
             */
            for (var day=0; day<7;++day) {
                // First cells of the previous month and the first days of next month
                if( ct < this.offset || ct > (this.days_in_month + this.offset - 1) ) {
                    str += "<td class=\"kal-inactive-cell kal-cell\">&nbsp;&nbsp;";
                    if ( ct > (this.days_in_month + this.offset - 1) ) {
                        str += ++day_of_next_month;
                    } else {
                        str += prev_day++;
                    }
                    str += "</td>";
                } else {
                    ++day_of_month; // Count the days of the month
                    // Get holidays and other events
                    var holiday = this.isEvent(this.year,this.month,day_of_month,day,"H");
                    var info = this.isEvent(this.year,this.month,day_of_month,day,"I");
                    var personal = this.isEvent(this.year,this.month,day_of_month,day,"P");
                    var text = "";
                    if ( holiday != "") {
                        text = "<br />&nbsp;&nbsp;" + holiday;
                    }
                    if ( info != "") {
                        text += "<br />&nbsp;&nbsp;" + info;
                    }
                    if ( personal != "") {
                        text += "<br />&nbsp;&nbsp;" + personal;
                    }
                    ////////////////////////////////////////////////////////////////////////////
                    /**
                     *  Set the cell type
                     */
                    var now = new Date();
                    if ( this.year == now.getFullYear() 
                      && this.month == now.getMonth() 
                      && day_of_month == now.getDate() ) {
                        str += "<td class=\"kal-today-cell kal-cell kal-comment-cell\">";
                    } else if ( holiday != "") {
                        str += "<td class=\"kal-holiday-cell kal-cell kal-comment-cell\">";
                    } else if ( info != "") {
                        str += "<td class=\"kal-info-cell kal-cell kal-comment-cell\">";
                    } else if ( personal != "") {
                        str += "<td class=\"kal-personal-cell kal-cell kal-comment-cell\">";
                    } else if ( day == 0 || day == 6 ) {
                        str += "<td class=\"kal-weekend-cell kal-cell kal-comment-cell\">";
                    } else {
                        str += "<td class=\"kal-base-cell kal-cell kal-comment-cell\">";
                    }
                    ////////////////////////////////////////////////////////////////////////////
                    /**
                     * Display the day of the month.
                     */
                    str += "<span style=\"font-size:1.3em\">&nbsp;&nbsp;" + day_of_month + "</span>";
                    ////////////////////////////////////////////////////////////////////////////
                    /**
                     *  Display phase of the Moon
                     */
                    switch(phase_days[day_of_month]) {
                        case this.NEW_MOON :
                            str += "&nbsp;&nbsp;" + "<img src=\"images/moon0.png\" width=\"20px\" height=\"20px\">";
                            break;
                        case this.FIRST_QUARTER :
                            str += "&nbsp;&nbsp;" + "<img src=\"images/moon2.png\" width=\"20px\" height=\"20px\">";
                            break;
                        case this.FULL_MOON :
                            str += "&nbsp;&nbsp;" + "<img src=\"images/moon4.png\" width=\"20px\" height=\"20px\">";
                            break;
                        case this.LAST_QUARTER :
                        str += "&nbsp;&nbsp;" + "<img src=\"images/moon6.png\" width=\"20px\" height=\"20px\">";
                            break;
                    }
                    ////////////////////////////////////////////////////////////////////////////
                    /**
                     * Display Content.
                     */
                    str += "<span class=\"kal-holiday-text\">" + text + "</span><br />";
                    for (var i=0;i<this.notesList.length; ++i) {
                        if ( this.year == this.notesList[i].year && this.month == (this.notesList[i].month-1) && day_of_month == this.notesList[i].day ) {
                            str += "<span class=\"kal-content\">" + "&nbsp;&nbsp;" + this.notesList[i].text + "</span><br />\n";
                        }
                    }
                    ////////////////////////////////////////////////////////////////////////////
                    /**
                     * Tool Tips - includes full data.
                     */
                    str += "<span class=\"kal-comment\">" + this.monthNames[this.month] + "&nbsp;" + day_of_month + "," + this.year + "<br />";
                    if (!isNaN(this.latitude)) {
                        var sundate = new Date(this.year,this.month,day_of_month,16,0,0);
                        var times = SunCalc.getTimes(sundate, this.latitude, this.longitude);
                        ////////////////////////////////////////////////////////////////////////
                        //
                        // Tool Tips - includes Sunrise & Sonset.
                        //
                        if ( times.sunrise.getMinutes() < 10 ) {
                            str += "Sunrise " + times.sunrise.getHours() + ":0" + times.sunrise.getMinutes();
                        } else {
                            str += "Sunrise " + times.sunrise.getHours() + ":" + times.sunrise.getMinutes();
                        }
                        str += "<br />";
                        if ( times.sunset.getMinutes() < 10 ) {
                            str += "Sunset " + times.sunset.getHours() + ":0" + times.sunset.getMinutes();
                        } else {
                            str += "Sunset " + times.sunset.getHours() + ":" + times.sunset.getMinutes();
                        }
                    }
                    // End of Tool tips.
                    str += "</span>";
                    // End of cell.
                    str += "</td>";
                }
                ++ct;    
            }
            str += "\n</tr>\n";
            $('#' + this.tab).append(str);
        }
    } //End of addRows()
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     *  Display the calendar.
     */
    static display() {
        // If the buttons have already been displayed simply update the months & year
        if ( this.displayed ) {
            $('#kal-month').text(this.monthNames[this.month]);
            $('#kal-prev-button').text(this.monthNames[this.prev_date.getMonth()]);
            $('#kal-year-button').text(this.year);
            $('#kal-next-button').text(this.monthNames[this.next_date.getMonth()]);
        } else {
            // If this is the first time, add the buttons
            this.displayed = true;
            $('#' + this.div).html("");
            $('#' + this.div).append("<p id=\"kal-month\" align=\"center\">" + this.monthNames[this.month] + "</p>");
            $('#' + this.div).append("<div id=\"kal-header\" align=\"center\">\n</div>");
            // previous month button
            $('#' + this.head).append("<button class=\"kal-button\" id=\"kal-prev-button\" onclick=\"Kalendar.prevMonth()\" align=\"center\">" + this.monthNames[this.prev_date.getMonth()] + "</button>");
            // previous year button
            $('#' + this.head).append("<button class=\"kal-year-button\" id=\"kal-prev-year-button\" onclick=\"Kalendar.prevYear()\" align=\"center\"><strong><</strong></button>");
            // display year button (may be used to display a full year in the future
            $('#' + this.head).append("<button class=\"kal-button\" id=\"kal-year-button\" align=\"center\" onclick='Kalendar.displayYear(Kalendar.year)'>" + this.year + "</button>");
            // next year button
            $('#' + this.head).append("<button class=\"kal-year-button\" id=\"kal-next-year-button\" onclick=\"Kalendar.nextYear()\" align=\"center\"><strong>></strong></button>");
            // next month button
            $('#' + this.head).append("<button class=\"kal-button\" id=\"kal-next-button\" onclick=\"Kalendar.nextMonth()\" align=\"center\">" + this.monthNames[this.next_date.getMonth()] + "</button>");
            $('#' + this.div).append("<br />");
            $('body').append($('<div/>', {
                id: 'kal-notes-dlog',
                title: 'Add Note'
            }));
            $('body').append($('<div/>', {
                id: 'kal-year-dlog'
            }));
        }
        $('#kal-table').html("");
        $('#' + this.div).append("<table id=\"kal-table\" align=\"center\">\n</table>");
        Kalendar.addDayHeader();
        Kalendar.addRows();
        // Initialize the calendar dialog box.
        if ( this.notesAllowed ) {
            $( "#" + this.div ).dialog({
                autoOpen: false,
                modal: true,
                open: function() {
                    $(this).css("height", "105ch");
                    $(this).parent().css("height", "auto");
                    $(this).parent().css("width", "auto");
                    $(this).parent().position({
                        my: "center",
                        at: "center",
                        of: window,
                        collision: "fit"
                    });
                },
                buttons: {
                    "TODAY": function() {
                        const today = new Date();
                        Kalendar.setDate(today.getFullYear(),today.getMonth()+1,today.getDate());
                    },
                    "ADD NOTE": function() {
                        Kalendar.addNotesTable();
                        $('#kal-notes-dlog').dialog('open');
                    },
                    "OK": function() {
                        $( this ).dialog( "close" );
                    }
                }
            });
        } else {
            $( "#" + this.div ).dialog({
                autoOpen: false,
                modal: true,
                open: function() {
                    $(this).css("height", "105ch");
                    $(this).parent().css("height", "auto");
                    $(this).parent().css("width", "auto");
                    $(this).parent().position({
                        my: "center",
                        at: "center",
                        of: window,
                        collision: "fit"
                    });
                },
                buttons: {
                    "TODAY": function() {
                        const today = new Date();
                        Kalendar.setDate(today.getFullYear(),today.getMonth()+1,today.getDate());
                    },
                    "OK": function() {
                        $( this ).dialog( "close" );
                    }
                }
            });
        }
        // Initialize the year dialog box.
        $( "#kal-year-dlog" ).dialog({
            autoOpen: false,
            modal: true,
            open: function() {
                $(this).parent().css("height", "auto");
                $(this).parent().css("width", "auto");
                $(this).parent().position({
                    my: "center",
                    at: "center",
                    of: window,
                    collision: "fit"
                });
            },
            buttons: {
                "OK": function() {
                    $( this ).dialog( "close" );
                },
            }
        });
        // Initialize the notes dialog box.
        $( "#kal-notes-dlog" ).dialog({
            autoOpen: false,
            height: "30em",
            modal: true,
            open: function() {
                // Set the inner content to fill the parent's height
                $(this).css("height", "30em");
                // Ensure the overall dialog container keeps the desired height
                $(this).parent().css("height", "auto");
                $(this).parent().css("width", "auto");
                $(this).parent().position({
                    my: "center",
                    at: "center",
                    of: window,
                    collision: "fit"
                });
            },
            buttons: {
                "CANCEL": function() {
                    $( this ).dialog( "close" );
                },
                "SAVE": function() {
                    $.post('saveNotes.php', { 'notes' : JSON.stringify(Kalendar.notesList) });
                    $( this ).dialog( "close" );
                    // Update the current calendar
                    $('#kal-table').html("");
                    $('#' + this.div).append("<table id=\"kal-table\" align=\"center\">\n</table>");
                    Kalendar.addDayHeader();
                    Kalendar.addRows();
                },
            }
        });
        $("#" + Kalendar.div ).css("height", "105ch");
        $("#" + Kalendar.div ).parent().css("height", "auto");
        $("#" + Kalendar.div ).parent().css("width", "auto");
        $("#" + Kalendar.div ).parent().position({
            my: "center",
            at: "center",
            of: window,
            collision: "fit"
        });
    } // End of display();
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     *  Allow Notes
     *
     *  @param {boolean} allow - Turn on or off adding notes
     */
    static allowNotes(allow) {
        Kalendar.notesAllowed = allow;
        Kalendar.display();
    } // End of allowNotes()
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     * Input the note list.
     *
     *  @param {string[]} list - The list of notes
     */
    static setNotesList(list) {
        Kalendar.notesList = list;
        Kalendar.display();
    } // End of setNotesList()
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     *  A note is a line of text that will be displayed in a given date cell. 
     */
    static addNote() {
        var dateStr = $('#datepicker').val();
        var textStr = $('#noteText').val();
        var parts = dateStr.split("-");
        var yearStr = parts[0];
        var mon = parts[1];
        var dy = parts[2];
        var new_note = { year: parts[0], month: parseInt(parts[1]), day: parseInt(parts[2]), text: textStr };
        this.notesList.push(new_note);
        this.addNotesTable();

    } // End of addNote()
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     *  The list of notes are stored in a JSON formatted file.
     */
    static readNotes() {
        $.post('readNotes.php',
            function(data) {
                Kalendar.setNotesList(JSON.parse(data));
            }, "json" );
    } // End of readNotes()
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     *  Add the content of the note dialog box.
     */
    static zeroPad = (num, places) => String(num).padStart(places, '0');
    static addNotesTable() {
        $('#kal-notes-dlog').html("");
        $('#kal-notes-dlog').append("<div id='kal-notes-header'/>");
        $('#kal-notes-header').append("<strong>Date: </strong><input type='text' id='datepicker' style='height: 18px'>");
        $('#kal-notes-header').append("&nbsp;&nbsp;&nbsp;&nbsp;");
        $('#kal-notes-header').append("<strong>Text: </strong><input type'text' id='noteText' size='30' style='height: 18px'>");
        $('#kal-notes-header').append("&nbsp;<button class='kal-add-button' id='addButton' style=\"border-radius: 5px\" onclick='Kalendar.addNote()'>ADD</button>");
        $('#kal-notes-dlog').append("<div id='kal-notes-div'/>");
        $('#kal-notes-div').append("<table>\n");
        $('#kal-notes-div').append("<tr><th style='width: 15em'>DATE</th><th style='width: 30em'>TEXT</th</tr>\n");
        var str = "";
        for ( var i=0; i<this.notesList.length; ++i) {
            str += "<tr>\n<td class=\"kal-note-table-cell\" style='width: 15em'>";
            str += this.notesList[i].year + "-" + this.zeroPad(this.notesList[i].month,2) + "-" + this.zeroPad(this.notesList[i].day,2);
            str += "</td><td align=\"left\" class=\"kal-note-table-cell\" style='width: 30em'>" + this.notesList[i].text + "</td>\n";
            str += "</td><td align=\"left\" class=\"kal-note-table-cell\" >";
            str += "<button class=\"kal-trash-button\" style=\"border-radius: 5px\" onclick='Kalendar.delNote(" + i + ")'><img id='kal-trash-icon' /></button>\n";
            str += "</td>\n</tr>\n";
        }
        $('#kal-notes-div').append(str);
        $('#kal-notes-div').append("</table>\n");
        $("#datepicker").datepicker({
            showButtonPanel: true,
            showOtherMonths: true,
            selectOtherMonths: true,
            changeMonth: true,
            changeYear: true,
            dateFormat: "yy-mm-dd",
            showOn: "button",
            buttonImage: "images/calendar2.png",
            buttonImageOnly: true,
            buttonText: "Select date"
        });
        $(".ui-datepicker-trigger").css({"height":"26px","width":"26px","top":"8px","left":"3px","position":"relative"});

    } // End of addNotesTable()
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     *  Delete a note from the current list of notes. 
     *  Note this does not delete the note from the file. The "SAVE" button in the note dialog
     *  box will rewrite the file.
     *
     *  @param {number} note - The index of the note to be deleted
     */
    static delNote(note) {
        this.notesList.splice(note,1);
        this.addNotesTable();
    } // End of delNote()
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     *  Display the entire year in a pop-up box. 
     *
     *  @param {number} year - The year to be displayed
     */
    static displayYear(year) {
        $('#kal-year-dlog').dialog({title:year});
        $('#kal-year-dlog').html("");
        $('#kal-year-dlog').append("<div id='kal-year-div' align='center' />");
        var str = "";
        str += "<table align='center'>";
        var month = 0;
        for ( var row=0;row<4;++row) {
            str += "<tr>";
            for ( var col=0;col<3;++col) {
                ++month;
                str += "<td id='kal-year-month" + month + "' class='kal-year-cell'>";
                str += "</td>";
            }
            str += "</tr>";
        }
        $('#kal-year-div').append(str);
        this.displayMonths(year);
        $('#kal-year-dlog').dialog("open");

    } // End of displayYear()
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     *  Display the months in the year pop-up box.
     *
     *  @param {number} year - The year to be displayed
     */
    static displayMonths(year) {
        var day = 0;
        var ct = 0;
        var str = "";
        for ( var month=1;month<=12;++month) {
            var date = new Date(year,month-1,1);
            var offset = date.getDay();
            var days_in_month = this.getDaysInMonth(year,month-1);
            str += "<table>";
            str += "<caption><strong>" + this.monthNames[month-1] + "</strong></caption>";
            for ( var dow=0;dow<7;++dow) {
                str += "<th class='kal-year-day-cell kal-header-color'>" + this.dayShortNames[dow] + "</th>";
            }
            //var num_rows = 5;
            //if ( offset + days_in_month > (5 * 7) ) {
            //    num_rows = 6;
            //}
            for ( var row=0;row<6;++row) {
                str += "<tr>";
                for ( var col=0;col<7;++col) {
                    ++ct;
                    if ( ct > offset && (ct - offset) <= days_in_month ) {
                        ++day;
                        var now = new Date();
                        var holiday = this.isEvent(year,month-1,day,col,"H");
                        var info = this.isEvent(year,month-1,day,col,"I");
                        var personal = this.isEvent(year,month-1,day,col,"P");
                        var text = "";
                        if ( holiday != "") {
                            text += "<span class='kal-comment'>" + holiday + " </span>";
                        }
                        if ( info != "") {
                            text += "<span class='kal-comment'>" + info + " </span>";
                        }
                        if ( personal != "") {
                            text += "<span class='kal-comment'>" + personal + " </span>";
                        }
                        if ( year == now.getFullYear() 
                          && month-1 == now.getMonth() 
                          && day == now.getDate() ) {
                            str += "<td class='kal-year-day-cell kal-today-color kal-comment-cell'>" + day + text + "</td>";
                        } else if ( holiday != "") {
                            str += "<td class='kal-year-day-cell kal-holiday-color kal-comment-cell'>" + day + text + "</td>";
                        } else if ( info != "") {
                            str += "<td class='kal-year-day-cell kal-info-color kal-comment-cell'>" + day + text + "</td>";
                        } else if ( personal != "") {
                            str += "<td class='kal-year-day-cell kal-parsonal-color kal-comment-cell'>" + day + text + "</td>";
                        } else if ( col == 0 || col == 6 ) {
                            str += "<td class='kal-year-day-cell kal-weekend-color kal-comment-cell'>" + day + text + "</td>";
                        } else {
                            str += "<td class='kal-year-day-cell kal-day-color kal-comment-cell'>" + day + text + "</td>";
                        }
                    } else {
                        str += "<td class='kal-year-day-cell kal-inactive-color'>&nbsp;</td>";
                    }
                }
                str += "</tr>";
            }
            $('#kal-year-month' + month).append(str);
            ct = 0;
            day = 0;
            str = "";
        }
    } // End of displayMonth()
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     * addEvents()
     * 
     * Add an event.
     *  
     *  Event structure "mm/dd/T:text"
     *   - mm              month 1-12
     *   - dd or position  day of month ( 1-31 or day position )
     *   - T               Event type "H" = Holiday, "I" = Information, "P" = Personal
     *   - text            Displayed text
     * 
     *  The day position:
     *   First 3 characters are the day of the week.
     *   Followed by the position first, second, third, fourth or last.
     *   The day position ignores case.
     *
     *  @param {string} ev - The event
     */
    static addEvent(ev) {
        this.events.push(ev);
        this.display();
    } // End of addEvents()
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /** 
     * isMatch ( Holidays, Information and Personal events )
     *
     *  @param {string} str         - The day of month of position 
     *  @param {number} day_of_week - The day of the week (0-6)
     *  @oaram {number} day         - The day of the month (1-31)
     */
    static dayShortNames = ['SUN','MON','TUE','WED','THU','FRI','SAT'];
    static isMatch(str,day_of_week,day) {

        var dowStr = str.substr(0,3);
        var week_day = -1;

        dowStr = dowStr.toUpperCase();
        for (var idx=0;idx<this.dayShortNames.length; ++idx) {
            if ( this.dayShortNames[idx] == dowStr ) {
                week_day = idx;
                break;
            }
        }
        if ( week_day == day_of_week ) {
            var posStr = str.substr(3,str.length-3);
            posStr = posStr.toUpperCase();
            switch(posStr) {
                case 'FIRST' :
                    if ( day <= 7 ) return true;
                    break;
                case 'SECOND' :
                    if ( day > 7 && day <= 14 ) return true;
                    break;
                case 'THIRD' :
                    if ( day > 14 && day <= 21 ) return true;
                    break;
                case 'FOURTH' :
                    if ( day > 21 && day <= 28 ) return true;
                    break;
                case 'LAST' :
                    if ( day > (this.days_in_month - 7) ) return true;
                    break;
            }
        }
        return false;

    } // End of isMatch()
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     * isEvent()
     *
     * Check if a data is a holiday. 
     *
     * @param {number} year         - The year
     * @param {number} month        - The month
     * @param {number} day          - The day of the month
     * @param {number} day_of_week  - The day of week
     * @param {string} type         - Type of event
     */
    static isEvent(year,month,day,day_of_week,type) {
        var rtn = "";
        var season = Date.getSeasons(year);
        const easter_date = this.easter(year);
        if ( easter_date.getMonth() == month && easter_date.getDate() == day && type == "H") {
            rtn += ("Easter");
        }
        ////////////////////////////////////////////////////////////////////////////////////////
        //
        // Display the Seasons
        //
        if ( month == 2 && season[1].getDate() == day && type == "I") {
            rtn += "Spring Equinox";
        }
        if ( month == 5 && season[2].getDate() == day && type == "I") {
            rtn += "Summer Solstice";
        }
        if ( month == 8 && season[3].getDate() == day && type == "I") {
            rtn += "Fall Equinox";
        }
        if ( month == 11 && season[4].getDate() == day && type == "I") {
            rtn += "Winter Solstice";
        }
        for (var i=0;i<this.events.length;++i) {
            var item = this.events[i].split(":");
            var when = item[0].split("/");
            var day_of_month = parseInt(when[1]);
            var event_month = parseInt(when[0]);
            if ( isNaN(day_of_month) && (event_month-1) == month && when[2] == type) {
                if ( this.isMatch(when[1],day_of_week,day) == true ) {
                    rtn += (item[1]);
                }
            }
            if ( (event_month-1) == month && day_of_month == day && when[2] == type ) {
                rtn += (item[1]);
            }
        }
        return rtn;
    } // End of isEvent()
    //
    ////////////////////////////////////////////////////////////////////////////////////////////
    /**
     * Calculate Easter
     *
     *  @param {number} year - The year
     */
    static easter(year) {
        var date, a, b, c, m, d;
        // Instantiate the date object.
        date = new Date;
        // Set the timestamp to midnight.
        date.setHours(0,0,0,0);
        // Set the year.
        date.setFullYear(year);
        // Find the golden number.
        a = year % 19;
        // Choose which version of the algorithm to use based on the given year.
        b = ( 2200 <= year && year <= 2299 ) ?
            ( ( 11 * a ) + 4 ) % 30 :
            ( ( 11 * a ) + 5 ) % 30;
        // Determine whether or not to compensate for the previous step.
        c = ( ( b === 0 ) || ( b === 1 && a > 10 ) ) ?
            ( b + 1 ) :
            b;
        // Use c first to find the month: April or March.
        m = ( 1 <= c && c <= 19 ) ? 3 : 2;
        // Then use c to find the full moon after the northward equinox.
        d = ( 50 - c ) % 31;
        // Mark the date of that full moon—the "Paschal" full moon.
        date.setMonth(m,d);
        // Count forward the number of days until the following Sunday (Easter).
        date.setMonth(m,d+(7-date.getDay()));
        // Gregorian Western Easter Sunday
        return date;
    } // End of easter()

} // End of class Kalendar

////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////// End of File //////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
