///////////////////////////////////////////////////////////////////////////////////////////////
//
//  moonphase.js
// 
// Written by: Don Dugger
//       Date: 2024-04-29
//
// Copyright (C) 2024 Acme Software Works, Inc.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
// GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
// OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
///////////////////////////////////////////////////////////////////////////////////////////////
/** @class */
class MoonPhase {

    ///////////////////////////////////////////////////////////////////////////////////////////////
    RAD = (3.1415926535897/180.0);
    IGREG = (15+31*(10+12*1582));
    MEAN_LUNAR_CYCLE = 29.53058868;
    ///////////////////////////////////////////////////////////////////////////////////////////////


    /**
     * flmoon() calculates the julian date of nth lunar cycle and nph phase
     * of the moon since Jan 1900
     * n is the number of lunar cycle since Jan 1900 - ???
     * This is the original function from the book.
     * It's been modified to work in JavaScript.
     * I also rewrote it to be litte clearer. And added
     * comments where I unterstood what was going on and
     * increased the accuracy of a few constants.
     * @param {number} n     The number of lunar cycles.
     * @param {number} nph   The lunar phase.
     * @return {number}     The full Julian date of the nth lunar cycle plus nph phase
     *             Note that the integer and fractional parts are
     *             returned by references.
     */
    flmoon(n,nph)
    {
        var jd = 0;
        var int_part = 0;
        var lunar_cycles = n + ( nph / 4.0 ); // the total number lunar cycle
                                              // nph = 4 is one lunar cycle 
        var t = lunar_cycles / 1236.85;       // 1236.85 is the number of lunar cycle per Julian Century
        var t2 = t * t;                       // Square for frequent use

        /** Sun's mean anomaly */
        var sun_anomaly = 359.2242 + ( 29.10535608 * lunar_cycles );

        /** Moon's mean anomaly */
        var moon_anomaly = 306.0253 + ( 385.81691806 * lunar_cycles ) + ( 0.010730 * t2 );

        /** Not sure of what's up here, but two of the numbers very interesting
         *  Notice that 2415020.75933 is used a referce epochs for mean time of lunar cycle which is 1900 Jan 1 at 6:13 AM
         *  And 29.53058868 is the "Synodic month" or the mean time in days between new moons.
         */
        var xtra = 0.75933 + ( 1.53058868 * lunar_cycles ) + ( ( ( 1.78e-4 ) - ( 1.55e-7 ) * t ) * t2 );

        /** Ok 2415020 is Julian date 1899 at noon + .75933 is the Julian date of a new moon
         *  28 + 1.53058868 is mean time of a lunar cycle ,
         *  and 7 close to 7.38264717 a lunar phase but unlike
         *  the first two the tractional part dose not seem to be accounted for. 
         */
        jd = 2415020 + ( 28 * n ) + ( 7 * nph );

        /** Looks like this is all being done to adjust the variations in the maen lunar cycle of 29.53058868 
         *  Which from what I understand is due the moons orbit in relationship to the earths orbit around the sun. 
         */
        if(nph == 0 || nph == 2) // New or Full - sun earth moon inline
            xtra += ( ( 0.1734-3.93e-4 * t ) * Math.sin(this.RAD*sun_anomaly) ) - ( 0.4068 * Math.sin(this.RAD*moon_anomaly) );
        else // 1st quarder last quarter - moon 90 deegree left or right of the earth and suns line 
            xtra += ( ( 0.1721-4.0e-4 * t ) * Math.sin(this.RAD*sun_anomaly) ) - ( 0.6280 * Math.sin(this.RAD*moon_anomaly) );

        /** This parts easy just put the integer and fractional parts right */
        int_part = parseInt( xtra >= 0.0 ? Math.floor(xtra) : Math.ceil(xtra-1.0) );
        jd += int_part;
        var frac = xtra - int_part;
        return (jd + frac);

    } // End of flmoon


    ///////////////////////////////////////////////////////////////
    /**
     * Julian2Calendar() computes the julian date
     * @param {number} jd    The Julian date number.
     * @param {number} mon   The month.
     * @param {number} day   The day of the month.
     * @param {number} year  The year.
     * @return {number}      Note that all the return values are references.
    */
    nextphase(year,mon,day,hr,min,nph)
    {
        var jd = 0;
        var djd = 0.0;
        var ph_jd = 0;
        var frac = 0.0;
        /** 12.37 is the number of lunar cycles per year
         *  n then is the number of lunar cycles since 1900 and the year month entered
         */
        var n = parseInt(12.37*((year-1900)+((mon-0.5)/12.0)));
        var intpart = 0.0;

        // Get the starting Julian date  */
        jd = this.Calendar2Julian(year,mon,day);
        djd = this.flmoon(n,nph) + 0.5;
        while(jd < djd) {
            djd = this.flmoon(--n,nph) + 0.5;
        }
        while(jd > djd) {
            djd = this.flmoon(++n,nph) + 0.5;
        }


        var now = new Date();
        djd -= (now.getTimezoneOffset() * (1.0/1440.0));
        if(this.isDST(now)) djd -= (1.0/24.0);
        ph_jd = parseInt(Math.floor(djd));
        if(this.isDST(this.Julian2Calendar(jd))) {
            djd += (1.0/24.0);
            ph_jd = parseInt(Math.floor(djd));
        }
        frac = djd - Math.floor(djd);
        
        var date = this.Julian2Calendar(ph_jd);
        frac *= 24.0;
        hr = parseInt(frac >= 0.0 ? Math.floor(frac) : Math.ceil(frac-1.0));
        frac -= hr;
        min = parseInt(Math.floor(60*frac));
        date.setHours(hr);
        date.setMinutes(min);
        return date;

    } // End of nextphase()

    ///////////////////////////////////////////////////////////////
    /** isDST() returns true if the given date is day light savings time
     *     There are two versions, one which takes the Julian date and
     *     one that takes the monuth, day and year.
     * @param {Date} date  The date
     * @return {boolean}
     */
    isDST(date) {
        let jan = new Date(date.getFullYear(), 0, 1).getTimezoneOffset();
        let jul = new Date(date.getFullYear(), 6, 1).getTimezoneOffset();
        return Math.max(jan, jul) !== date.getTimezoneOffset();    
    }

    ///////////////////////////////////////////////////////////////
    /** Julian2Calendar() computes the julian date
     *  @param {number} jd    The Julian date number.
     *  @param {number} mon   The month.
     *  @param {number} day   The day of the month.
     *  @param {number} year  The year.
     *  @return  Note that all the return values are references.
     */
    Julian2Calendar(jd)
    {
        const GREG = 2299161;
        var a = 0;
        var b = 0;
        var c = 0;
        var d = 0;
        var e = 0;
        var year = 0;
        var month = 0;
        var day = 0;

        if(jd >= GREG) {
            a = parseInt((((jd-1867216)-0.25)/36524.25));
            a = parseInt(jd+1+parseInt(a)-parseInt(0.25*parseInt(a)));
        } else {
            a = jd;
        }
        b = a+1524;
        c = parseInt(6680+((b-2439870)-122.1)/365.25);
        d = 365*c+parseInt(0.25*c);
        e = parseInt((b-d)/30.6001);
        day = parseInt((b-d-parseInt(30.6001*e)));
        month = parseInt(e-1);
        if(month > 12) month -= 12;
        year = parseInt(c-4715);
        if(month > 2) --year;
        if(year <= 0) --year;
        return new Date(year,month-1,day);

    } // End of Julian2Calendar

    //////////////////////////////////////////////////////////////
    /** calendar2Julian() computes the date given a julian date
     * @param {number} mon   The month.
     * @param {number} day   The day of the month.
     * @param {number} year  The year.
     */
    Calendar2Julian(year,month,day)
    {
        var jul;
        var ja,jy=year,jm;

        if(jy == 0) return(0);
        if(jy < 0) ++jy;
        if(month > 2) jm=month+1;
        else {
            --jy;
            jm=month+13;
        }
        jul = parseInt(Math.floor(365.25*jy)+Math.floor(30.6001*jm)+day+1720995);
        if(day+31*(month+12*year) >= this.IGREG) {
            ja=parseInt(0.01*jy);
            jul += 2-ja+parseInt(0.25*ja);
        }
        return(jul);

    } // End of Calendar2Julian

} // End of class MoonPhase

///////////////////////////////////////////////////////////////
///////////////////// END OF FILE /////////////////////////////
///////////////////////////////////////////////////////////////
